/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#ifndef H_SHUNTING
#define H_SHUNTING

#include "base.h"
#include "tokener.h"

#include <stack>
#include <list>

using namespace std;

struct op_stub {
  opid op;
  int index;
};

#define NODE_NONE 0
#define NODE_OP   1
#define NODE_VAL  2

//#define UDEF_COPY_CONSTRUCTORS

#define SPLICE_ONE
#define SPLICE_TWO
#define SPLICE_LIST

#define TRANSFER_ONE
#define TRANSFER_TWO
#define TRANSFER_LIST

struct treenode {
  uint type;
  uint opcode;
  value val;
#if DEBUG>1
  const char* opname;
#endif
  list<treenode> childz;
  treenode(){
    type=NODE_NONE;
  }
  treenode(const opid& o){
    type=NODE_OP;
    opcode=o.code;
#if DEBUG>1
    opname=o.name;
#endif
  }
  treenode(const value& v){
    type=NODE_VAL;
    val=v;
  }
  /*
  void set_op(const opid& o){
    type=NODE_OP;
    opcode=o.code;
#if DEBUG>1
    opname=o.name;
#endif
  }
  void set_val(value v){
    type=NODE_VAL;
    val=v;
  }
  */
#ifdef UDEF_COPY_CONSTRUCTORS
  treenode(const treenode& tn);
  treenode& operator=(const treenode& tn);
#endif
  #if DEBUG>1
  void treeplot(FILE* fajl);
  #endif
};

class compiler {
 private:
  stack<op_stub> op_stack;
  list<treenode> out_stack;


  void stack_putNONE(const opid& op);
  void stack_putONE(const opid& op);
  void stack_putTWO(const opid& op);


  void push_op(const opid& op);
  void push_par(uint type);
  void pop_par(uint type);
  void push_val(const value& v);
  void shift_op();
  uchar top_precedence();
  int is_overflown();
  void process(const token& t);

 public:
  compiler(char* expression, treenode& output);
};


#endif
