/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#ifndef H_VMACHINE
#define H_VMACHINE

#include "base.h"
#include "shunting.h"
#include "unroll.h"

#include <vector>
#include <map>
#include <stack>

#ifdef HAVE_LIBREADLINE
#  if defined(HAVE_READLINE_READLINE_H)
#    include <readline/readline.h>
#  elif defined(HAVE_READLINE_H)
#    include <readline.h>
#  else
#    undef HAVE_LIBREADLINE
#  endif
#endif
#ifdef HAVE_READLINE_HISTORY
#  if defined(HAVE_READLINE_HISTORY_H)
#    include <readline/history.h>
#  elif defined(HAVE_HISTORY_H)
#    include <history.h>
#  else
#    undef HAVE_READLINE_HISTORY
#  endif
#endif

#define BUF_SIZE 4096

using namespace std;


/*
 * representation of string slice
 * c: 's': misc string;
 *    'd': number;
 */
struct format_piece{
  char c; 	// c: type of input field
  char* parsed_start; // actual input field (parsed according to c)
  char* parsed_stop; // end of input field (where 0 should be)
};

#define XL_LINE 0
#define XL_EXTENDED 1

struct code_chunk {
  treenode code;
  bool printable;
};

typedef map<int,array> t_mem;
typedef map<int,treenode> t_fun;
typedef stack<array> t_stack;

class eval_control {
 private:
#ifndef NOSYSCALL
  clock_t start_time;
#endif

  int funcount;
  t_fun funspace;
  t_stack argstack;
  int* functrl;
  int maxit;

  t_mem memory;
  int memroot;
  int memdefault;
  int mempos;

  int toggle_history;
  int histsize;//stores the number of history entries

  int toggle_print;

  char awaiting;
  char* lfile;
  int xltoggle;

  FILE* outstream;

  char* delim;
  char* Fd;
  char* Ff;

  int bufsize;
  char* buffer;//buffer for reading files
  char* bufgets(FILE* file);

  void syscall(array& a, array& out);
  void evaluate_operator(uint code, int argc, array* args, array& out);
  int evaluate(treenode& node, array& out);
  void evaluate_code(list<code_chunk>& code, bool printlast, array& out);
  void interlace_code(list<code_chunk>& code);
  void print_array(array& a, FILE* stream, const char* sep, char end);
  void memory_dump();
  void credits_dump();
  void configure_dump();
  FILE* input_file_open(char* filename);
  FILE* output_file_open(char* filename);
  int precompile(char* line, char** comment,list<code_chunk>& code);
  char* strip(char* str);
  char* formatdup(const char* fmt);

  int retrieve_code(char* f, list<code_chunk>& code, bool file_or_line);

  int format_splitline(char* line, vector<format_piece>& fmt);

  void add_to_history(array& arr);

 public:
  void evaluate_line(char* line, char** comment, bool printlast, array& out);

  void read_interactive();
  void read_file(FILE* f);

  void load_file(FILE* f);

  void interlace_to_line(char* line);
  void interlace_to_file(char* f, FILE* l);

  void apply_to_file(char* f, FILE* l, bool file_or_line);
  void extended_to_file(char* l, FILE* f, bool file_or_line);

  void input_line(char* line);
  eval_control();
  ~eval_control();
};


#endif
